/*
 * 
 * Best code so far. 10 bit fractional frequency counter gives smallest
 * mains sync tracking errors
 * 
 * No more Bessel LP filter of Vfb. It runs fine with the
 * nanoverter Vfb opamp LP filter in place
 */

#define ADC_OUT  0          // channel for AC output sample

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
//#define PPWM  665       // 60.062 Hz - choose one or the other.
#define PPWM  799         // 50.002 Hz 799  49.939 Hz for 800

      
uint8_t uf,oen,v1low, lv_stopped,int_f;
uint16_t pcount;
uint32_t vpwr;
int sst;
uint16_t l[NPWM+1];
uint16_t ml[NPWM+1];
float pwr;
uint8_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float xv[5],yv[5],sm_pca;
float ac_output, ac_setpoint;

void init_vars(void);

int pcount_acc,new_delta_int;
int pcount_delta, phase_error, old_phase_error;
int pcint,new_delta, p_lock, sdcount=0;

uint8_t init_done; 
uint8_t tt=0;

void setup()
  {
  int i;
  float t,u;

 
  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   
    u = 65500.0 * (0.98*sin(t));   // 1/2 wave sine lookup table, scaled to (16 bits - a small amount)
    l[i] =  (int)u;  
    }
 
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(12, OUTPUT);    // inverter ON/OFF LED, pin 15 on nano, Arduino D12
  pinMode(13,OUTPUT);
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  pinMode(5,OUTPUT);      // used for IR2110 shutdown drive, LOW = shutdown
  pinMode(7,OUTPUT);      // DSO sync pulse, each 50Hz
  pinMode(2,INPUT_PULLUP);
  noInterrupts();

  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);

  EICRA = 3;               // interrupt on falling edge from pin D2
  EIMSK |= bit (INT0);     // enable it  
  cbi(PORTD,4);   // no sync
  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.
  oen = 0;        // output enable. = 1 to run inverter, = 0 to stop inverter
  init_vars();
  p_lock = 0;          // phase lock flag = 0 = NOT locked
  pcount_acc = 0;      // fractional integer math. Each PWM, pcount_delta is added to pcount_acc and divided by 1024
  pcount_delta = 1024; // in effect we are getting an addition of 1. But pcount_delta can be modified, a little bigger or smaller..
  old_phase_error = 0; // used in PID, to obtain derivative
  init_done = 1;
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  
  
  cbi(PORTB,5);    // D13, LED
  Serial.begin(115200);     // remove in prodduction code
  }

//
//
ISR (INT0_vect)
  {
  int bd = 500;
  int pe_frac=0,tc;
  tc = TCNT1;
  if  (int_f == 0) 
    return;  // only once per 50 Hz cycle. Noise! much noise...
  int_f = 0;

// best tracking accuracy yet. need integral term now due to different feedback
// into 20kHz timer clock
//

    pe_frac = (32*tc)/PPWM;
    pcount_acc = 0;
    phase_error = 32*(pcount - 32 +  (v1low == 1?200:0) - 200) + pe_frac;                                                                                  
    new_delta = 2*phase_error ;
    new_delta_int += new_delta;
    pcount_acc = -new_delta - new_delta_int/4 ;
  
  if (abs(phase_error) < 50)
    {
      p_lock = 1;      sbi(PORTD,4);     sbi(PORTB,5);              
    }
  else
    {
      p_lock=0;      cbi(PORTD,4);      cbi(PORTB,5);
    }
  }  

//
// 20Khz SPWM, code has only 50uS to run, takes about 10uS
//
ISR(TIMER1_OVF_vect)        
 {   
  long c;
  int i;
  //
  // alter the 20kHz PWM rate in accordance with PLL loop correction
  // if pcount_delta = 0 then no change needed
  // so this is the amount needed to steer the output back into phase lock
  // It is not the error of lock, it is the amount needed to change from 20kHz
  //
  i = pcount_delta/256;
  ICR1 = PPWM - i;
  pcount_delta = (pcount_delta - 256 * i) + pcount_acc + 135;       // 135 is trim for the 16MHz clock is not quite 16.0000
  //
  // look up PWM duty for this part of the 1/2 cycle 
  c = (l[pcount] * vpwr) >> 16;   // scale sine wave by vpwr, 32 bit integer calcs. 16 bit shift fastest of all
  if (v1low == 1)                 // drive only this pin
    OCR1A = PPWM/2 + c;                    
  else
    OCR1A = PPWM/2 - c;                    
  
  // just go to next duty % lookup, 0 to 199 inclusive (compare with previous code)
  pcount ++; 
  
  if(pcount >= NPWM)          // pcount  will = NPWM, at each 1/2 wave, at the start
    {
    pcount = 0;               // reset counter
    uf=1;                     // enable one PID control loop execution, in loop()
    if (v1low == 1)           // if first 1/2 wave..
      {
       sbi(PORTD,7);     //DSO pulse to tell mains sync pulse occurred in maybe wrong time
       v1low = 0;             // toggle to 2nd half wave
                                // int_f = 1 enables INT0 to execute, which sets int_f = 0 if run. If not run, it remains = 1
       if (int_f == 1)          // no AC sync signal at all, pull sync lock output LOW. 
        {                       
          cbi(PORTD,4);         // tell world, no sync
          pcount_delta = 0;   // after sync loss, also when/if no sync lock ever attained, ensure AC output is 50Hz
        }
       int_f = 1;               // this is here, last to execute, to ensure this PWM code runs first and is not delayed by
      }                         // subsequent AC sync interrupt processing. 
    else
      {
       v1low = 1;             // toggle to 1st half wave
       cbi(PORTD,7); 
      }
    }
}


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  ICR1 = PPWM;
  pcount_acc=0;
  lv_stopped = 0;  
  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr= 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;  // debounce counter for input switch 
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;        // initialise PID
  lastErr = 0.0;
  int_f = 0;           //used in AC sync, one shot per 50Hz
  if (p_lock == 1)
    {
     // nothing to do so far 
    }
  else
    { 
    p_lock = 0;          // phase lock flag = 0 = NOT locked
    pcount_delta = 0; // in effect we are getting an addition of 1. But pcount_delta can be modified, a little bigger or smaller..
    old_phase_error = 0; // used in PID, to obtain derivative
    }
  cbi(PORTD,4);  
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}


void loop()
{
  float ch0;

  // sample channel..
  // In the past, I low pass filtered this. not now, no need thanks to active LP filter on Vfb
  ac_output = (float)analogRead(ADC_OUT) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
 
  if(uf == 1  )    // this will execute at 100Hz
      {            // it runs at near zero volt crossing of AC output       
      uf=0; 
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        {
        sst = 0;
        cbi(PORTD,5);     // pull IR2184 shutdown LOW to disable gate drive output
        }
      else
        sbi(PORTD,5);     // pull it HIGH, to enable output

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
          else                    // During slow start need to ensure output is tracking setpoint
            pwr -= 1.0/250.0;     // should output get close enough to matter.
          }
        else
          pwr -= 1.0 /250.0;      // this will execute in slow stop
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        }
        
      if (pwr > 0.99) pwr = 0.99;     // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01;     // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)((float)(PPWM/2) * pwr);// finally apply required PWM duty power factor to vpwr     
      check_switch_cont(); 
      
      sm_pca = (float)pcount_acc;
      xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4];
  xv[4] = sm_pca / 2.259747797e+05;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4];
  yv[4] =   (xv[0] + xv[4]) + 4 * (xv[1] + xv[3]) + 6 * xv[2]
                     + ( -0.7428578687 * yv[0]) + (  3.1954036268 * yv[1])
                     + ( -5.1599230905 * yv[2]) + (  3.7073065280 * yv[3]);
  sm_pca = yv[4];

      if (sdcount++ > 50)
        {
        sdcount=0;  
        Serial.print(sm_pca);
        Serial.print(" ");
        //Serial.print(phase_error);
        Serial.println(" -400 400");
        }
      }

}

void check_switch_cont()
{   // called at 100 Hz
  if ((PINB & 0x01) == 0x01)     
    {
    if (init_done == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,4); // and light LED on D12, pin 15 of nano
    }

  // scr shutdown - without a dummy connection of D6 to ground, inverter wont run.
  if (((PINB & 0x01) == 0) || ((PIND & 0x40) == 0x40)) 
  // for testing, do not check scr shutdown pin, just nano2 on/off
  //if (((PINB & 0x01) == 0) )       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,4); // and blank LED on D12, pin 15 of nano
    }
          
}
